import os
import posixpath
from typing import Self

# TODO: Cut any dependency of DocumentNode on the backend implementation
#       Maybe use some king of "yield" to achieve this?


class DocumentNode():
    def __init__(self, name: str, children: list):
        self.name = name
        self.children = children

    def print_tree(self, level=0) -> None:
        print('-' * level + self.name)

        for child in self.children:
            child.print_tree(level + 1)

    def get_child(self, path: str) -> Self | None:
        # TODO: Handle paths like "a//b"
        # TODO: Handle paths like "/"

        # Stop condition, we found it
        if path == '':
            return self

        # Go recursively through each link in the chain
        components = path.split('/')
        for child in self.children:
            if child.name == components[0]:
                # If there's just one left '' will be sent
                subpath = '/'.join(components[1:])
                return child.get_child(subpath)

        # A link in the chain doesn't exist
        return None

    @staticmethod
    def get_tree(base: str, path: str) -> "DocumentNode":
        children = list()

        # Split and remerge for Windows
        internal_path = os.path.join(base, *path.split('/'))

        if not os.path.exists(internal_path):
            raise Exception("Tried to open a folder that doesn't exist")

        _, dirs, files = next(os.walk(internal_path))

        for dir in dirs:
            subpath = posixpath.join(path, dir)
            children.append(DocumentNode.get_tree(base, subpath))

        # Second pass, add the files. If it's a .md file, strip the extension,
        # checking for conflicts with folders
        #
        # TODO: Maybe this isn't the best way to define what the text of a
        #       folder should be.
        #
        # TODO: What if there's a file obscuring a folder?
        #       Check if the superposing is a .md file, warn the user otherwise
        for file in files:
            if file.endswith('.md'):
                file = file[:-3]

            if file not in children:
                children.append(DocumentNode(file, []))

        # With this algorithm they're not properly sorted
        children.sort(key=lambda x: x.name)

        return DocumentNode(path.split('/')[-1], children)


class DocumentTree():
    def __init__(self, base: str):
        self.base = base
        self.root = DocumentNode.get_tree(base, '')

    def get_node(self, path: str) -> DocumentNode | None:
        return self.root.get_child(path)

    def reload(self) -> bool:
        # TODO: Instead of doing this, actually reload
        self.root = DocumentNode.get_tree(self.base, '')
        return True

    # ----- Document Operations -----

    # Split and remerge for Windows
    def document_read(self, path: str) -> str | None:
        internal_path = os.path.join(self.base, *path.split('/')) + '.md'
        if os.path.exists(internal_path):
            return open(internal_path, 'r').read()

        # File doesn't exist
        return None

    def document_write(self, path: str, text: str) -> None:
        internal_path = os.path.join(self.base, *path.split('/')) + '.md'

        # TODO: Probably add some sort of confirmation beforehand?
        os.makedirs(os.path.dirname(internal_path), exist_ok=True)
        with open(internal_path, 'w') as file:
            file.write(text)

        self.reload()

    # -------------------------------
